/*
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * SPDX-FileCopyrightText: Copyright (C) 2020 Jacob Barkdull
 */


#ifndef _GSM_SETAFFINITY_H_
#define _GSM_SETAFFINITY_H_

#include <glib.h>
#include "application.h"

void create_set_affinity_dialog (GsmApplication *app);

#endif /* _GSM_SETAFFINITY_H_ */
